/*
 * linux/arch/arm/mach-psionw/time.c
 *
 * RTC handler for Psion 5mx.
 *
 * Copyright (C) 2001 Yuji Shinokawa <ysh@mob.or.jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/hardware/psionw.h>

#undef DEBUG_RTC

/*
 * Offset to compatible with EPOC time handling.
 * In EPOC, it seems a time for rtc = 0 is "00:00:00 1 Jan 2000 UTC"
 * and rtc has signed value.
 *
 *     RTC         Time & date
 * 0x00000000 = 00:00:00  1 Jan 2000 (London Inner, United Kingdom)
 *
 * You can get this offset by following perl script;
 *   #!/usr/bin/perl
 *   require 'timelocal.pl';
 *   $time = timegm (0, 0, 0, 1, 0, 2000);
 *   print "$time\n";
 */
#define EPOC_RTC_OFFSET		946684800

extern int (*set_rtc) (void);
extern unsigned long (*gettimeoffset) (void);

unsigned int psionw_read_rtc(void)
{
	int try_count, rtc1, rtc2;
	u_int xrtc;

	try_count = 3;
	do {
		rtc1 =
		    (psionw_readl(RTCDRU) << 16) | (psionw_readl(RTCDRL) &
						    0xffff);
		rtc2 =
		    (psionw_readl(RTCDRU) << 16) | (psionw_readl(RTCDRL) &
						    0xffff);
	} while (rtc1 != rtc2 && --try_count);

#if DEBUG_RTC
	printk("DEBUG: (psionw_get_rtc) read rtc = %08x\n", rtc1);
#endif

	xrtc = rtc1 + EPOC_RTC_OFFSET;

	return xrtc;
}

void psionw_write_rtc(u_int xrtc)
{
	int rtc;

	rtc = xrtc - EPOC_RTC_OFFSET;

#if DEBUG_RTC
	printk("DEBUG: (psionw_set_rtc) write rtc = %08x\n", rtc);
#endif

	/* Writing order is important.  */
	psionw_writel((rtc >> 16), RTCDRU);
	psionw_writel((rtc & 0xffff), RTCDRL);
}

unsigned int psionw_read_rtc_alarm(void)
{
	int try_count, rtc1, rtc2;
	u_int xrtc;

	try_count = 3;
	do {
		rtc1 =
		    (psionw_readl(RTCMRU) << 16) | (psionw_readl(RTCMRL) &
						    0xffff);
		rtc2 =
		    (psionw_readl(RTCMRU) << 16) | (psionw_readl(RTCMRL) &
						    0xffff);
	} while (rtc1 != rtc2 && --try_count);

#if DEBUG_RTC
	printk("DEBUG: (psionw_get_rtc_alarm) read rtc alarm = %08x\n", rtc1);
#endif

	xrtc = rtc1 + EPOC_RTC_OFFSET;

	return xrtc;
}

void psionw_write_rtc_alarm(u_int xrtc)
{
	int rtc;

	rtc = xrtc - EPOC_RTC_OFFSET;

#if DEBUG_RTC
	printk("DEBUG: (psionw_set_rtc_alarm) write rtc alarm = %08x\n", rtc);
#endif

	/* Writing order is important.  */
	psionw_writel((rtc >> 16), RTCMRU);
	psionw_writel((rtc & 0xffff), RTCMRL);
}

EXPORT_SYMBOL(psionw_read_rtc);
EXPORT_SYMBOL(psionw_write_rtc);
EXPORT_SYMBOL(psionw_read_rtc_alarm);
EXPORT_SYMBOL(psionw_write_rtc_alarm);

static int
psionw_set_rtc(void)
{
	u_int xrtc, rtc;

	xrtc = (u_int) xtime.tv_sec;

	psionw_write_rtc(xrtc);

	rtc = psionw_read_rtc();
	if (xrtc > rtc || rtc - xrtc > 1)
		return 1;	/* Write failed.  */

	return 0;
}

static unsigned long
psionw_gettimeoffset(void)
{
	return 0;
}

static int
psionw_rtc_init(void)
{
	/* Set the initial RTC value.  */
	xtime.tv_sec = (time_t) psionw_read_rtc();

	/* Setup hook.  */
	set_rtc = psionw_set_rtc;
	gettimeoffset = psionw_gettimeoffset;

	return 0;
}

__initcall(psionw_rtc_init);
